import random
import time

import pyray
import raylib

from classField import Field
from classWall import *


class Entity:

    def __init__(self, image_path: str, velocity, pos):
        self.pos = pos
        self.velocity = velocity
        self.texture: rl.Texture = ...
        self.set_texture(image_path)
        self.color = rl.WHITE

    def set_texture(self, image_path):
        image = rl.load_image(image_path)
        self.texture = rl.load_texture_from_image(image)
        rl.unload_image(image)

    def check_wall(self, field: Field):
        return isinstance(
            field.field[
                (self.pos[1] + self.velocity[1]) % Field.height
            ][
                (self.pos[0] + self.velocity[0]) % Field.width
            ],
            Wall)

    def draw(self):
        rl.draw_texture(
            self.texture,
            Field.padding_x + 20 * (self.pos[0]) - 1,
            Field.padding_y + 20 * (self.pos[1]) - 1,
            self.color
        )

    def move(self):
        self.pos[0] = (self.pos[0] + self.velocity[0]) % Field.width
        self.pos[1] = (self.pos[1] + self.velocity[1]) % Field.height


class Ghost(Entity):
    def __init__(self, pos, color):
        super().__init__('./images/ghost.png', [0, -1], pos)
        self.color = color
        self.init_color = color
        self.initial_pos = [*pos]
        self.dead = False
        self.dead_start = None
    
    def kill(self):
        self.dead = True
        self.dead_start = time.time()
        self.color = rl.GRAY
    
    def revive(self):
        self.dead = False
        self.dead_start = None
        self.color = self.init_color
        self.pos = self.initial_pos

    def process(self, field: Field):
        if not self.dead:
            self.velocity = random.choice([
                [1, 0],
                [-1, 0],
                [0, 1],
                [0, -1],
                [0, 0]
            ])
            if not self.check_wall(field):
                self.move()

        # whether dead frames should end
        now = time.time()
        if self.dead and now - self.dead_start > 15:
            self.revive()


class ManPac(Entity):
    def __init__(self, pos: list[int]):
        super().__init__(
            image_path='images/Pacman-right.png',
            velocity=[0, 0],
            pos=pos
        )
        self.color = rl.YELLOW
        self.super = False
        self.super_start = None
        self.lives = 3
        self.invincible = False
        self.invincible_start = None
        self.score = 0
        self.seeds = 512

    def handle_wasd(self):
        if rl.is_key_pressed(rl.KeyboardKey.KEY_W):
            self.velocity = [0, -1]
            self.set_texture('images/Pacman-up.png')
        elif rl.is_key_pressed(rl.KeyboardKey.KEY_S):
            self.velocity = [0, 1]
            self.set_texture('images/Pacman-down.png')
        elif rl.is_key_pressed(rl.KeyboardKey.KEY_D):
            self.velocity = [1, 0]
            self.set_texture('images/Pacman-right.png')
        elif rl.is_key_pressed(rl.KeyboardKey.KEY_A):
            self.velocity = [-1, 0]
            self.set_texture('images/Pacman-left.png')

    def check_seed(self, field: Field):
        return isinstance(
            field.field[self.pos[1]][self.pos[0]],
            Seed
        )

    def check_big_seed(self, field: Field):
        return isinstance(
            field.field[self.pos[1]][self.pos[0]],
            BigSeed
        )

    def check_ghosts(self, ghosts: list[Ghost]) -> bool | int:
        for (index, ghost) in enumerate(ghosts):
            if ghost.dead:
                continue
            if self.pos[0] == ghost.pos[0] and self.pos[1] == ghost.pos[1]:
                return index
        return False

    def gain_invincibility(self):
        self.lives -= 1
        self.invincible = True
        self.invincible_start = time.time()
        self.color = rl.RED

    def lose_invincibility(self):
        self.invincible = False
        self.invincible_start = None
        if self.color == rl.RED:
            self.color = rl.YELLOW

    def gain_super(self):
        self.super = True
        self.super_start = time.time()
        self.color = rl.BLUE

    def lose_super(self):
        self.super = False
        self.super_start = None
        self.color = rl.YELLOW

    def process(self, field: Field, ghosts: list[Ghost]):

        # check wall
        if self.check_wall(field):
            self.velocity = [0, 0]

        self.move()

        # check seeds
        if self.check_seed(field):
            field.field[self.pos[1]][self.pos[0]] = 'nothing'
            self.score += 10
            self.seeds -= 1
        elif self.check_big_seed(field):
            field.field[self.pos[1]][self.pos[0]] = 'nothing'
            self.score += 20
            self.seeds -= 1
            self.gain_super()

        # check ghosts
        if index := self.check_ghosts(ghosts):
            if not self.super and not self.invincible:
                self.gain_invincibility()
            elif self.super:
                self.score += 50
                ghosts[index].kill()

        # whether super and invincible states should end
        now = time.time()
        if self.super and now - self.super_start > 15:
            self.lose_super()
        if self.invincible and now - self.invincible_start > 4:
            self.lose_invincibility()
